"""
This module defines the main visitor algorithm transforming ufl expressions
to pymbolic and loopy.
"""
from dune.perftool.error import PerftoolUFLError
from dune.perftool.generation import (domain,
                                      get_global_context_value,
                                      )
from dune.perftool.ufl.flatoperators import get_operands
from dune.perftool.ufl.modified_terminals import (ModifiedTerminalTracker,
                                                  Restriction,
                                                  )
from dune.perftool.ufl.execution import Expression
from dune.perftool.tools import maybe_wrap_subscript
from loopy import Reduction

from pymbolic.primitives import (Call,
                                 Product,
                                 Quotient,
                                 Subscript,
                                 Sum,
                                 Variable,
                                 )

from ufl.algorithms import MultiFunction
from ufl.checks import is_cellwise_constant
from ufl.functionview import select_subelement
from ufl import (VectorElement,
                 TensorElement,
                 )
from ufl.classes import (FixedIndex,
                         IndexSum,
                         JacobianDeterminant,
                         )


class UFL2LoopyVisitor(ModifiedTerminalTracker):
    def __init__(self, interface, measure, dimension_indices):
        self.interface = interface
        self.measure = measure
        self.dimension_indices = dimension_indices

        # Call base class constructors
        super(UFL2LoopyVisitor, self).__init__()

    # Allow recursion through self.call(..)
    call = MultiFunction.__call__

    def __call__(self, o):
        # Reset some state variables that are reinitialized for each accumulation term
        self.indices = None
        self.inames = ()

        return self.call(o)

    #
    # Form argument/coefficients handlers:
    # This is where the actual domain specific work happens
    #

    def argument(self, o):
        # Correct the restriction on boundary integrals
        restriction = self.restriction
        if self.measure == 'exterior_facet':
            restriction = Restriction.NEGATIVE

        # Select the correct subtree of the finite element
        element = select_subelement(o.ufl_element(), self.component)
        leaf_element = element

        indices = self.indices
        if indices is None:
            indices = ()
        self.indices = None

        # Now treat the case of this being a vector finite element
        if element.num_sub_elements() > 0:
            # I cannot handle general mixed elements here...
            assert isinstance(element, (VectorElement, TensorElement))

            # If this is a vector element, we need add an additional accumulation loop iname
            shape = len(element.value_shape())
            indices = indices[shape:]
            for i in range(len(element.value_shape())):
                self.inames = self.inames + (self.interface.dimension_iname(context='arg', count=i),)

            # For the purpose of basis evaluation, we need to take the leaf element
            leaf_element = element.sub_elements()[0]

        if self.grad:
            raise PerftoolUFLError("Gradients should have been transformed to reference gradients!!!")

        # Have the issued instruction depend on the iname for this localfunction space
        inames = self.interface.lfs_inames(leaf_element, restriction, o.number())
        self.inames = self.inames + inames

        if self.reference_grad:
            return maybe_wrap_subscript(self.interface.pymbolic_reference_gradient(leaf_element, restriction, o.number()), indices)
        else:
            return maybe_wrap_subscript(self.interface.pymbolic_basis(leaf_element, restriction, o.number()), indices)

    def coefficient(self, o):
        # Do something different for trial function and coefficients from jacobian apply
        if o.count() == 0 or o.count() == 1:
            # Correct the restriction on boundary integrals
            restriction = self.restriction
            if self.measure == 'exterior_facet':
                restriction = Restriction.NEGATIVE

            if self.grad:
                raise PerftoolUFLError("Gradients should have been transformed to reference gradients!!!")

            if self.reference_grad:
                if o.count() == 0:
                    return self.interface.pymbolic_trialfunction_gradient(o.ufl_element(), restriction, self.component)
                else:
                    return self.interface.pymbolic_apply_function_gradient(o.ufl_element(), restriction, self.component)
            else:
                if o.count() == 0:
                    return self.interface.pymbolic_trialfunction(o.ufl_element(), restriction, self.component)
                else:
                    return self.interface.pymbolic_apply_function(o.ufl_element(), restriction, self.component)

        # Check if this is a parameter function
        else:
            # We expect all coefficients to be of type Expression!
            assert isinstance(o, Expression)

            # Determine the name of the parameter function
            name = get_global_context_value("data").object_names[id(o)]

            cellwise_constant = is_cellwise_constant(o)

            # Trigger the generation of code for this thing in the parameter class
            if o.on_intersection:
                self.interface.intersection_parameter_function(name, o, cellwise_constant)
            else:
                self.interface.cell_parameter_function(name, o, self.restriction, cellwise_constant)

            # And return a symbol
            return Variable(name)

    #
    # Handlers for all indexing related stuff
    #

    def indexed(self, o):
        # Handle indices first to allow the aggregate handler to mangle in indices
        # This is necessary in some places where the locality of the tree transformation
        # is not given (easiest example: jacobian_inverse handler, complex example:
        # sum factorziation handler for trial function gradient)
        self.indices = self.call(o.ufl_operands[1])

        # Handle the aggregate!
        aggr = self.call(o.ufl_operands[0])

        # self.indices being None means they are already handled during recursion!
        if self.indices is None:
            return aggr
        else:
            indices = self.indices
            self.indices = None
            return maybe_wrap_subscript(aggr, indices)

    def index_sum(self, o, additional_inames=()):
        # There is three scenarios here:
        # * This is a normal IndexSum:
        #   We should collect the reduction inames, go into recursion and issue
        #   a reduction operation to a temporary
        # * This is part of a nested indexsum:
        #   We only collect the reduction iname and go into recursion
        # * This IndexSum is implicitly handled by the accumulation process
        #   We only go into recursion

        # Get the iname for the reduction index
        ind = o.ufl_operands[1][0]
        redinames = additional_inames + (ind,)
        shape = o.ufl_operands[0].ufl_index_dimensions[0]
        domain(self.interface.name_index(ind), shape)

        # If the left operand is an index sum to, we do it in one reduction
        if isinstance(o.ufl_operands[0], IndexSum):
            return self.index_sum(o.ufl_operands[0], additional_inames=redinames)
        else:
            # Recurse to get the summation expression
            term = self.call(o.ufl_operands[0])
            redinames = tuple(i for i in redinames if i not in self.dimension_indices)
            if len(redinames) > 0:
                ret = Reduction("sum", tuple(self.interface.name_index(ind) for ind in redinames), term)
            else:
                ret = term

            return ret

    def _index_or_fixed_index(self, index):
        if isinstance(index, FixedIndex):
            return index._value
        else:
            if index in self.dimension_indices:
                self.inames = self.inames + (self.dimension_indices[index],)
                return Variable(self.dimension_indices[index])
            else:
                return Variable(self.interface.name_index(index))

    def multi_index(self, o):
        return tuple(self._index_or_fixed_index(i) for i in o)

    def index(self, o):
        return self._index_or_fixed_index(o)

    #
    # Handlers for arithmetic operators and functions
    # Those handlers would be valid in any code going from UFL to pymbolic
    #

    def product(self, o):
        return Product(tuple(self.call(op) for op in get_operands(o)))

    def float_value(self, o):
        return o.value()

    def int_value(self, o):
        return o.value()

    def division(self, o):
        return Quotient(self.call(o.ufl_operands[0]), self.call(o.ufl_operands[1]))

    def sum(self, o):
        return Sum(tuple(self.call(op) for op in get_operands(o)))

    def zero(self, o):
        return 0

    def abs(self, o):
        if isinstance(o.ufl_operands[0], JacobianDeterminant):
            return self.call(o.ufl_operands[0])
        else:
            return Call('abs', self.call(o.ufl_operands[0]))

    #
    # Handlers for geometric quantities
    #

    def facet_normal(self, o):
        # The normal must be restricted to be well-defined
        assert self.restriction is not Restriction.NONE

        if self.restriction == Restriction.POSITIVE:
            return Variable(self.interface.name_unit_outer_normal())
        if self.restriction == Restriction.NEGATIVE:
            # It is highly unnatural to have this generator function,
            # but I do run into subtle trouble with return -1*outer
            # as the indexing into the normal happens only later.
            # Not investing more time into this cornercase right now.
            return Variable(self.interface.name_unit_inner_normal())

    def quadrature_weight(self, o):
        return self.interface.pymbolic_quadrature_weight()

    def jacobian_determinant(self, o):
        return Variable(self.interface.name_jacobian_determinant())

    def jacobian_inverse(self, o):
        restriction = self.restriction
        if self.measure == 'exterior_facet':
            restriction = Restriction.NEGATIVE

        # Dune only hast JacobianInverseTransposed as a first class citizen,
        # so we need to switch the indices around!
        assert(len(self.indices) == 2)
        indices = (self.indices[1], self.indices[0])
        self.indices = None

        return Subscript(Variable(self.interface.name_jacobian_inverse_transposed(restriction)), indices)

    def jacobian(self, o):
        raise NotImplementedError("How did you get Jacobian into your form? We only support JacobianInverse right now. Report!")

    def facet_jacobian_determinant(self, o):
        return Variable(self.interface.name_facet_jacobian_determinant())
